@@ -1,5 +1,5 @@
-#define version_info "(0,9,2,'final',1)"
-#define __version__ "0.9.2"
+#define version_info "(1,1,6,'final',1)"
+#define __version__ "1.1.6"
 /*
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -32,10 +32,10 @@ PERFORMANCE OF THIS SOFTWARE.
 
 #ifdef MS_WIN32
 #include <windows.h>
+#endif /* MS_WIN32 */
 #ifndef uint
 #define uint unsigned int
 #endif
-#endif /* MS_WIN32 */
 
 #include "structmember.h"
 #include "mysql.h"
@@ -75,7 +75,7 @@ typedef struct {
 	PyObject *converter;
 } _mysql_ConnectionObject;
 
-#define check_connection(c) if (!(c->open)) _mysql_Exception(c)
+#define check_connection(c) if (!(c->open)) return _mysql_Exception(c)
 #define result_connection(r) ((_mysql_ConnectionObject *)r->conn)
 #define check_result_connection(r) check_connection(result_connection(r))
 
@@ -92,6 +92,13 @@ typedef struct {
 
 extern PyTypeObject _mysql_ResultObject_Type;
 
+static int _mysql_server_init_done = 0;
+#if MYSQL_VERSION_ID >= 40000
+#define check_server_init(x) if (!_mysql_server_init_done) { if (mysql_server_init(0, NULL, NULL)) { _mysql_Exception(NULL); return x; } else { _mysql_server_init_done = 1;} }
+#else
+#define check_server_init(x) if (!_mysql_server_init_done) _mysql_server_init_done = 1
+#endif
+
 PyObject *
 _mysql_Exception(_mysql_ConnectionObject *c)
 {
@@ -99,10 +106,10 @@ _mysql_Exception(_mysql_ConnectionObject *c)
 	int merr;
 
 	if (!(t = PyTuple_New(2))) return NULL;
-	if (!(c->open)) {
+	if (!_mysql_server_init_done) {
 		e = _mysql_InternalError;
 		PyTuple_SET_ITEM(t, 0, PyInt_FromLong(-1L));
-		PyTuple_SET_ITEM(t, 1, PyString_FromString("connection is closed"));
+		PyTuple_SET_ITEM(t, 1, PyString_FromString("server not initialized"));
 		PyErr_SetObject(e, t);
 		Py_DECREF(t);
 		return NULL;
@@ -161,7 +168,120 @@ _mysql_Exception(_mysql_ConnectionObject *c)
 	Py_DECREF(t);
 	return NULL;
 }
+	  
+static char _mysql_server_init__doc__[] =
+"Initialize embedded server. If this client is not linked against\n\
+the embedded server library, this function does nothing.\n\
+\n\
+args -- sequence of command-line arguments\n\
+groups -- sequence of groups to use in defaults files\n\
+";
 
+static PyObject *_mysql_server_init(
+	PyObject *self,
+	PyObject *args,
+	PyObject *kwargs) {
+	static char *kwlist[] = {"args", "groups", NULL};
+	char **cmd_args_c=NULL, **groups_c=NULL, *s;
+	int cmd_argc=0, i, groupc;
+	PyObject *cmd_args=NULL, *groups=NULL, *ret=NULL, *item;
+
+	if (_mysql_server_init_done) {
+		PyErr_SetString(_mysql_ProgrammingError,
+				"already initialized");
+		return NULL;
+	}
+	  
+	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|OO", kwlist,
+					 &cmd_args, &groups))
+		return NULL;
+
+#if MYSQL_VERSION_ID >= 40000
+	if (cmd_args) {
+		if (!PySequence_Check(cmd_args)) {
+			PyErr_SetString(PyExc_TypeError,
+					"args must be a sequence");
+			goto finish;
+		}
+		cmd_argc = PySequence_Size(cmd_args);
+		if (cmd_argc == -1) {
+			PyErr_SetString(PyExc_TypeError,
+					"args could not be sized");
+			goto finish;
+		}
+		cmd_args_c = (char **) PyMem_Malloc(cmd_argc*sizeof(char *));
+		for (i=0; i< cmd_argc; i++) {
+			item = PySequence_GetItem(cmd_args, i);
+			s = PyString_AsString(item);
+			Py_DECREF(item);
+			if (!s) {
+				PyErr_SetString(PyExc_TypeError,
+						"args must contain strings");
+				goto finish;
+			}
+			cmd_args_c[i] = s;
+		}
+	}
+	if (groups) {
+		if (!PySequence_Check(groups)) {
+			PyErr_SetString(PyExc_TypeError,
+					"groups must be a sequence");
+			goto finish;
+		}
+		groupc = PySequence_Size(groups);
+		if (groupc == -1) {
+			PyErr_SetString(PyExc_TypeError,
+					"groups could not be sized");
+			goto finish;
+		}
+		groups_c = (char **) PyMem_Malloc((1+groupc)*sizeof(char *));
+		for (i=0; i< groupc; i++) {
+			item = PySequence_GetItem(groups, i);
+			s = PyString_AsString(item);
+			Py_DECREF(item);
+			if (!s) {
+				PyErr_SetString(PyExc_TypeError,
+						"groups must contain strings");
+				goto finish;
+			}
+			groups_c[i] = s;
+		}
+		groups_c[groupc] = (char *)NULL;
+	}
+	/* even though this may block, don't give up the interpreter lock
+	   so that the server can't be initialized multiple times. */
+	if (mysql_server_init(cmd_argc, cmd_args_c, groups_c)) {
+		_mysql_Exception(NULL);
+		goto finish;
+	}
+#endif
+	ret = Py_None;
+	Py_INCREF(Py_None);
+	_mysql_server_init_done = 1;
+  finish:
+	PyMem_Free(groups_c);
+	PyMem_Free(cmd_args_c);
+	return ret;
+}
+
+static char _mysql_server_end__doc__[] =
+"Shut down embedded server. If not using an embedded server, this\n\
+does nothing.";
+
+static PyObject *_mysql_server_end(
+	PyObject *self,
+	PyObject *args) {
+	if (_mysql_server_init_done) {
+#if MYSQL_VERSION_ID >= 40000
+		mysql_server_end();
+#endif
+		_mysql_server_init_done = 0;
+		Py_INCREF(Py_None);
+		return Py_None;
+	}
+	return _mysql_Exception(NULL);
+}
+	 
 #if MYSQL_VERSION_ID >= 32314
 static char _mysql_thread_safe__doc__[] =
 "Indicates whether the client is compiled as thread-safe.";
@@ -170,7 +290,8 @@ static PyObject *_mysql_thread_safe(
 	PyObject *self,
 	PyObject *args) {
 	PyObject *flag;
-	if (!PyArg_NoArgs(args)) return NULL;
+	if (!PyArg_ParseTuple(args, "")) return NULL;
+	check_server_init(NULL);
 	if (!(flag=PyInt_FromLong((long)mysql_thread_safe()))) return NULL;
 	return flag;
 }
@@ -233,6 +354,38 @@ _mysql_ResultObject_Initialize(
 			fun = Py_None;
 			Py_INCREF(Py_None);
 		}
+		if (PySequence_Check(fun)) {
+			int j, n2=PySequence_Size(fun);
+			PyObject *fun2=NULL;
+			for (j=0; j<n2; j++) {
+				PyObject *t = PySequence_GetItem(fun, j);
+				if (!t) continue;
+				if (!PyTuple_Check(t)) goto cleanup;
+				if (PyTuple_GET_SIZE(t) == 2) {
+					long mask;
+					PyObject *pmask=NULL;
+					pmask = PyTuple_GET_ITEM(t, 0);
+					fun2 = PyTuple_GET_ITEM(t, 1);
+					if (PyInt_Check(pmask)) {
+						mask = PyInt_AS_LONG(pmask);
+						if (mask & fields[i].flags) {
+							break;
+						}
+						else {
+							continue;
+						}
+					} else {
+						break;
+					}
+				}
+			  cleanup:
+				Py_DECREF(t);
+			}
+			if (!fun2) fun2 = Py_None;
+			Py_INCREF(fun2);
+			Py_DECREF(fun);
+			fun = fun2;
+		}
 		PyTuple_SET_ITEM(self->converter, i, fun);
 	}
 	return 0;
@@ -270,8 +423,14 @@ _mysql_ConnectionObject_Initialize(
 	PyObject *args,
 	PyObject *kwargs)
 {
-	MYSQL *conn;
+	MYSQL *conn=NULL;
 	PyObject *conv = NULL;
+	PyObject *ssl = NULL;
+	PyObject *value = NULL;
+#if HAVE_OPENSSL
+	char *key = NULL, *cert = NULL, *ca = NULL,
+		*capath = NULL, *cipher = NULL;
+#endif
 	char *host = NULL, *user = NULL, *passwd = NULL,
 		*db = NULL, *unix_socket = NULL;
 	uint port = MYSQL_PORT;
@@ -281,6 +440,7 @@ _mysql_ConnectionObject_Initialize(
 				  "connect_timeout", "compress",
 				  "named_pipe", "init_command",
 				  "read_default_file", "read_default_group",
+				  "client_flag", "ssl",
 				  NULL } ;
 	int connect_timeout = 0;
 	int compress = -1, named_pipe = -1;
@@ -290,14 +450,16 @@ _mysql_ConnectionObject_Initialize(
 	
 	self->converter = NULL;
 	self->open = 0;
-	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|ssssisOiiisss:connect",
+	check_server_init(-1);
+	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|ssssisOiiisssiO:connect",
 					 kwlist,
 					 &host, &user, &passwd, &db,
 					 &port, &unix_socket, &conv,
 					 &connect_timeout,
 					 &compress, &named_pipe,
 					 &init_command, &read_default_file,
-					 &read_default_group))
+					 &read_default_group,
+					 &client_flag, &ssl))
 		return -1;
 
 	if (!conv) 
@@ -310,7 +472,24 @@ _mysql_ConnectionObject_Initialize(
 		return -1;
 	self->converter = conv;
 
-	self->open = 1;
+#define _stringsuck(d,t,s) {t=PyMapping_GetItemString(s,#d);\
+        if(t){d=PyString_AsString(t);Py_DECREF(t);}\
+        PyErr_Clear();}
+	
+	if (ssl) {
+#if HAVE_OPENSSL
+		_stringsuck(ca, value, ssl);
+		_stringsuck(capath, value, ssl);
+		_stringsuck(cert, value, ssl);
+		_stringsuck(key, value, ssl);
+		_stringsuck(cipher, value, ssl);
+#else
+		PyErr_SetString(_mysql_NotSupportedError,
+				"client library does not have SSL support");
+		return -1;
+#endif
+	}
+
 	Py_BEGIN_ALLOW_THREADS ;
 	conn = mysql_init(&(self->connection));
 	if (connect_timeout) {
@@ -330,8 +509,16 @@ _mysql_ConnectionObject_Initialize(
 		mysql_options(&(self->connection), MYSQL_READ_DEFAULT_FILE, read_default_file);
 	if (read_default_group != NULL)
 		mysql_options(&(self->connection), MYSQL_READ_DEFAULT_GROUP, read_default_group);
+
+#if HAVE_OPENSSL
+	if (ssl)
+		mysql_ssl_set(&(self->connection),
+			      key, cert, ca, capath, cipher);
+#endif
+
 	conn = mysql_real_connect(&(self->connection), host, user, passwd, db,
 				  port, unix_socket, client_flag);
+  error:
 	Py_END_ALLOW_THREADS ;
 
 	if (!conn) {
@@ -344,6 +531,7 @@ _mysql_ConnectionObject_Initialize(
 	  be done here. tp_dealloc still needs to call PyObject_GC_UnTrack(),
 	  however.
 	*/
+	self->open = 1;
 	return 0;
 }
 
@@ -414,7 +602,8 @@ _mysql_ConnectionObject_close(
 	_mysql_ConnectionObject *self,
 	PyObject *args)
 {
-	if (!PyArg_NoArgs(args)) return NULL;
+	if (!args) return NULL;
+	if (!PyArg_ParseTuple(args, "")) return NULL;
 	if (self->open) {
 		Py_BEGIN_ALLOW_THREADS
 		mysql_close(&(self->connection));
@@ -436,7 +625,7 @@ _mysql_ConnectionObject_affected_rows(
 	_mysql_ConnectionObject *self,
 	PyObject *args)
 {
-	if (!PyArg_NoArgs(args)) return NULL;
+	if (!PyArg_ParseTuple(args, "")) return NULL;
 	check_connection(self);
 	return PyLong_FromUnsignedLongLong(mysql_affected_rows(&(self->connection)));
 }
@@ -471,7 +660,7 @@ _mysql_ConnectionObject_dump_debug_info(
 	PyObject *args)
 {
 	int err;
-	if (!PyArg_NoArgs(args)) return NULL;
+	if (!PyArg_ParseTuple(args, "")) return NULL;
 	check_connection(self);
 	Py_BEGIN_ALLOW_THREADS
 	err = mysql_dump_debug_info(&(self->connection));
@@ -481,6 +670,62 @@ _mysql_ConnectionObject_dump_debug_info(
 	return Py_None;
 }
 
+#if MYSQL_VERSION_ID >= 40100
+static char _mysql_ConnectionObject_autocommit__doc__[] =
+"Set the autocommit mode. True values enable; False value disable.\n\
+";
+static PyObject *
+_mysql_ConnectionObject_autocommit(
+	_mysql_ConnectionObject *self,
+	PyObject *args)
+{
+	int flag, err;
+	if (!PyArg_ParseTuple(args, "i", &flag)) return NULL;
+	Py_BEGIN_ALLOW_THREADS
+	err = mysql_autocommit(&(self->connection), flag);
+	Py_END_ALLOW_THREADS
+	if (err) return _mysql_Exception(self);
+	Py_INCREF(Py_None);
+	return Py_None;
+}		
+
+static char _mysql_ConnectionObject_commit__doc__[] =
+"Commits the current transaction\n\
+";
+static PyObject *
+_mysql_ConnectionObject_commit(
+	_mysql_ConnectionObject *self,
+	PyObject *args)
+{
+	int err;
+	if (!PyArg_ParseTuple(args, "")) return NULL;
+	Py_BEGIN_ALLOW_THREADS
+	err = mysql_commit(&(self->connection));
+	Py_END_ALLOW_THREADS
+	if (err) return _mysql_Exception(self);
+	Py_INCREF(Py_None);
+	return Py_None;
+}		
+
+static char _mysql_ConnectionObject_rollback__doc__[] =
+"Rolls backs the current transaction\n\
+";
+static PyObject *
+_mysql_ConnectionObject_rollback(
+	_mysql_ConnectionObject *self,
+	PyObject *args)
+{
+	int err;
+	if (!PyArg_ParseTuple(args, "")) return NULL;
+	Py_BEGIN_ALLOW_THREADS
+	err = mysql_rollback(&(self->connection));
+	Py_END_ALLOW_THREADS
+	if (err) return _mysql_Exception(self);
+	Py_INCREF(Py_None);
+	return Py_None;
+}		
+#endif
+
 static char _mysql_ConnectionObject_errno__doc__[] =
 "Returns the error code for the most recently invoked API function\n\
 that can succeed or fail. A return value of zero means that no error\n\
@@ -492,7 +737,7 @@ _mysql_ConnectionObject_errno(
 	_mysql_ConnectionObject *self,
 	PyObject *args)
 {
-	if (!PyArg_NoArgs(args)) return NULL;
+	if (!PyArg_ParseTuple(args, "")) return NULL;
 	check_connection(self);
 	return PyInt_FromLong((long)mysql_errno(&(self->connection)));
 }
@@ -508,7 +753,7 @@ _mysql_ConnectionObject_error(
 	_mysql_ConnectionObject *self,
 	PyObject *args)
 {
-	if (!PyArg_NoArgs(args)) return NULL;
+	if (!PyArg_ParseTuple(args, "")) return NULL;
 	check_connection(self);
 	return PyString_FromString(mysql_error(&(self->connection)));
 }
@@ -536,6 +781,7 @@ _mysql_escape_string(
 #if MYSQL_VERSION_ID < 32321
 	len = mysql_escape_string(out, in, size);
 #else
+	check_server_init(NULL);
 	if (self) {
 		check_connection(self);
 		len = mysql_real_escape_string(&(self->connection), out, in, size);
@@ -576,6 +822,7 @@ _mysql_string_literal(
 #if MYSQL_VERSION_ID < 32321
 	len = mysql_escape_string(out+1, in, size);
 #else
+	check_server_init(NULL);
 	if (self) {
 		check_connection(self);
 		len = mysql_real_escape_string(&(self->connection), out+1, in, size);
@@ -725,7 +972,7 @@ _mysql_ResultObject_describe(
 	PyObject *d;
 	MYSQL_FIELD *fields;
 	unsigned int i, n;
-	if (!PyArg_NoArgs(args)) return NULL;
+	if (!PyArg_ParseTuple(args, "")) return NULL;
 	check_result_connection(self);
 	n = mysql_num_fields(self->result);
 	fields = mysql_fetch_fields(self->result);
@@ -761,7 +1008,7 @@ _mysql_ResultObject_field_flags(
 	PyObject *d;
 	MYSQL_FIELD *fields;
 	unsigned int i, n;
-	if (!PyArg_NoArgs(args)) return NULL;
+	if (!PyArg_ParseTuple(args, "")) return NULL;
 	check_result_connection(self);
 	n = mysql_num_fields(self->result);
 	fields = mysql_fetch_fields(self->result);
@@ -1066,7 +1313,7 @@ _mysql_ConnectionObject_character_set_name(
 	PyObject *args)
 {
 	const char *s;
-	if (!PyArg_NoArgs(args)) return NULL;
+	if (!PyArg_ParseTuple(args, "")) return NULL;
 	check_connection(self);
 #if MYSQL_VERSION_ID >= 32321
 	s = mysql_character_set_name(&(self->connection));
@@ -1084,7 +1331,8 @@ _mysql_get_client_info(
 	PyObject *self,
 	PyObject *args)
 {
-	if (!PyArg_NoArgs(args)) return NULL;
+	if (!PyArg_ParseTuple(args, "")) return NULL;
+	check_server_init(NULL);
 	return PyString_FromString(mysql_get_client_info());
 }
 
@@ -1098,7 +1346,7 @@ _mysql_ConnectionObject_get_host_info(
 	_mysql_ConnectionObject *self,
 	PyObject *args)
 {
-	if (!PyArg_NoArgs(args)) return NULL;
+	if (!PyArg_ParseTuple(args, "")) return NULL;
 	check_connection(self);
 	return PyString_FromString(mysql_get_host_info(&(self->connection)));
 }
@@ -1113,7 +1361,7 @@ _mysql_ConnectionObject_get_proto_info(
 	_mysql_ConnectionObject *self,
 	PyObject *args)
 {
-	if (!PyArg_NoArgs(args)) return NULL;
+	if (!PyArg_ParseTuple(args, "")) return NULL;
 	check_connection(self);
 	return PyInt_FromLong((long)mysql_get_proto_info(&(self->connection)));
 }
@@ -1128,7 +1376,7 @@ _mysql_ConnectionObject_get_server_info(
 	_mysql_ConnectionObject *self,
 	PyObject *args)
 {
-	if (!PyArg_NoArgs(args)) return NULL;
+	if (!PyArg_ParseTuple(args, "")) return NULL;
 	check_connection(self);
 	return PyString_FromString(mysql_get_server_info(&(self->connection)));
 }
@@ -1144,8 +1392,8 @@ _mysql_ConnectionObject_info(
 	_mysql_ConnectionObject *self,
 	PyObject *args)
 {
-	char *s;
-	if (!PyArg_NoArgs(args)) return NULL;
+	const char *s;
+	if (!PyArg_ParseTuple(args, "")) return NULL;
 	check_connection(self);
 	s = mysql_info(&(self->connection));
 	if (s) return PyString_FromString(s);
@@ -1180,7 +1428,7 @@ _mysql_ConnectionObject_insert_id(
 	PyObject *args)
 {
 	my_ulonglong r;
-	if (!PyArg_NoArgs(args)) return NULL;
+	if (!PyArg_ParseTuple(args, "")) return NULL;
 	check_connection(self);
 	Py_BEGIN_ALLOW_THREADS
 	r = mysql_insert_id(&(self->connection));
@@ -1220,7 +1468,7 @@ _mysql_ConnectionObject_field_count(
 	_mysql_ConnectionObject *self,
 	PyObject *args)
 {
-	if (!PyArg_NoArgs(args)) return NULL;
+	if (!PyArg_ParseTuple(args, "")) return NULL;
 	check_connection(self);
 #if MYSQL_VERSION_ID < 32224
 	return PyInt_FromLong((long)mysql_num_fields(&(self->connection)));
@@ -1237,7 +1485,7 @@ _mysql_ResultObject_num_fields(
 	_mysql_ResultObject *self,
 	PyObject *args)
 {
-	if (!PyArg_NoArgs(args)) return NULL;
+	if (!PyArg_ParseTuple(args, "")) return NULL;
 	check_result_connection(self);
 	return PyInt_FromLong((long)mysql_num_fields(self->result));
 }	
@@ -1253,7 +1501,7 @@ _mysql_ResultObject_num_rows(
 	_mysql_ResultObject *self,
 	PyObject *args)
 {
-	if (!PyArg_NoArgs(args)) return NULL;
+	if (!PyArg_ParseTuple(args, "")) return NULL;
 	check_result_connection(self);
 	return PyLong_FromUnsignedLongLong(mysql_num_rows(self->result));
 }	
@@ -1276,7 +1524,7 @@ _mysql_ConnectionObject_ping(
 	PyObject *args)
 {
 	int r;
-	if (!PyArg_NoArgs(args)) return NULL;
+	if (!PyArg_ParseTuple(args, "")) return NULL;
 	check_connection(self);
 	Py_BEGIN_ALLOW_THREADS
 	r = mysql_ping(&(self->connection));
@@ -1341,7 +1589,7 @@ _mysql_ConnectionObject_select_db(
 
 static char _mysql_ConnectionObject_shutdown__doc__[] =
 "Asks the database server to shut down. The connected user must\n\
-have shutdown privileges. Non-stadard.\n\
+have shutdown privileges. Non-standard.\n\
 ";
 
 static PyObject *
@@ -1350,10 +1598,14 @@ _mysql_ConnectionObject_shutdown(
 	PyObject *args)
 {
 	int r;
-	if (!PyArg_NoArgs(args)) return NULL;
+	if (!PyArg_ParseTuple(args, "")) return NULL;
 	check_connection(self);
 	Py_BEGIN_ALLOW_THREADS
-	r = mysql_shutdown(&(self->connection));
+	r = mysql_shutdown(&(self->connection)
+#if MYSQL_VERSION_ID >= 40103
+		, SHUTDOWN_DEFAULT
+#endif
+		);
 	Py_END_ALLOW_THREADS
 	if (r) return _mysql_Exception(self);
 	Py_INCREF(Py_None);
@@ -1372,8 +1624,8 @@ _mysql_ConnectionObject_stat(
 	_mysql_ConnectionObject *self,
 	PyObject *args)
 {
-	char *s;
-	if (!PyArg_NoArgs(args)) return NULL;
+	const char *s;
+	if (!PyArg_ParseTuple(args, "")) return NULL;
 	check_connection(self);
 	Py_BEGIN_ALLOW_THREADS
 	s = mysql_stat(&(self->connection));
@@ -1396,7 +1648,7 @@ _mysql_ConnectionObject_store_result(
 	PyObject *arglist=NULL, *kwarglist=NULL, *result=NULL;
 	_mysql_ResultObject *r=NULL;
 
-	if (!PyArg_NoArgs(args)) return NULL;
+	if (!PyArg_ParseTuple(args, "")) return NULL;
 	check_connection(self);
 	arglist = Py_BuildValue("(OiO)", self, 0, self->converter);
 	if (!arglist) goto error;
@@ -1404,9 +1656,9 @@ _mysql_ConnectionObject_store_result(
 	if (!kwarglist) goto error;
 	r = MyAlloc(_mysql_ResultObject, _mysql_ResultObject_Type);
 	if (!r) goto error;
-	result = (PyObject *) r;
 	if (_mysql_ResultObject_Initialize(r, arglist, kwarglist))
 		goto error;
+	result = (PyObject *) r;
 	if (!(r->result)) {
 		Py_DECREF(result);
 		Py_INCREF(Py_None);
@@ -1435,7 +1687,7 @@ _mysql_ConnectionObject_thread_id(
 	PyObject *args)
 {
 	unsigned long pid;
-	if (!PyArg_NoArgs(args)) return NULL;
+	if (!PyArg_ParseTuple(args, "")) return NULL;
 	check_connection(self);
 	Py_BEGIN_ALLOW_THREADS
 	pid = mysql_thread_id(&(self->connection));
@@ -1457,7 +1709,7 @@ _mysql_ConnectionObject_use_result(
 	PyObject *arglist=NULL, *kwarglist=NULL, *result=NULL;
 	_mysql_ResultObject *r=NULL;
 
-	if (!PyArg_NoArgs(args)) return NULL;
+	if (!PyArg_ParseTuple(args, "")) return NULL;
 	check_connection(self);
 	arglist = Py_BuildValue("(OiO)", self, 1, self->converter);
 	if (!arglist) return NULL;
@@ -1548,7 +1800,7 @@ _mysql_ResultObject_row_tell(
 	PyObject *args)
 {
 	MYSQL_ROW_OFFSET r;
-	if (!PyArg_NoArgs(args)) return NULL;
+	if (!PyArg_ParseTuple(args, "")) return NULL;
 	check_result_connection(self);
 	r = mysql_row_tell(self->result);
 	return PyInt_FromLong(r-self->result->data->data);
@@ -1578,9 +1830,29 @@ static PyMethodDef _mysql_ConnectionObject_methods[] = {
 	{
 		"affected_rows",
 		(PyCFunction)_mysql_ConnectionObject_affected_rows,
-		0,
+		METH_VARARGS,
 		_mysql_ConnectionObject_affected_rows__doc__
 	},
+#if MYSQL_VERSION_ID >= 40100
+	{
+		"autocommit",
+		(PyCFunction)_mysql_ConnectionObject_autocommit,
+		METH_VARARGS,
+		_mysql_ConnectionObject_autocommit__doc__
+	},
+	{
+		"commit",
+		(PyCFunction)_mysql_ConnectionObject_commit,
+		METH_VARARGS,
+		_mysql_ConnectionObject_commit__doc__
+	},
+	{
+		"rollback",
+		(PyCFunction)_mysql_ConnectionObject_rollback,
+		METH_VARARGS,
+		_mysql_ConnectionObject_rollback__doc__
+	},
+#endif
 #if MYSQL_VERSION_ID >= 32303
 	{
 		"change_user",
@@ -1592,13 +1864,13 @@ static PyMethodDef _mysql_ConnectionObject_methods[] = {
 	{
 		"character_set_name",
 		(PyCFunction)_mysql_ConnectionObject_character_set_name,
-		0,
+		METH_VARARGS,
 		_mysql_ConnectionObject_character_set_name__doc__
 	},
 	{
 		"close",
 		(PyCFunction)_mysql_ConnectionObject_close,
-		0,
+		METH_VARARGS,
 		_mysql_ConnectionObject_close__doc__
 	},
 	{
@@ -1622,49 +1894,49 @@ static PyMethodDef _mysql_ConnectionObject_methods[] = {
 	{
 		"error",
 		(PyCFunction)_mysql_ConnectionObject_error,
-		0,
+		METH_VARARGS,
 		_mysql_ConnectionObject_error__doc__
 	},
 	{
 		"errno",
 		(PyCFunction)_mysql_ConnectionObject_errno,
-		0,
+		METH_VARARGS,
 		_mysql_ConnectionObject_errno__doc__
 	},
 	{
 		"field_count",
 		(PyCFunction)_mysql_ConnectionObject_field_count,
-		0,
+		METH_VARARGS,
 		_mysql_ConnectionObject_field_count__doc__
 	}, 
 	{
 		"get_host_info",
 		(PyCFunction)_mysql_ConnectionObject_get_host_info,
-		0,
+		METH_VARARGS,
 		_mysql_ConnectionObject_get_host_info__doc__
 	},
 	{
 		"get_proto_info",
 		(PyCFunction)_mysql_ConnectionObject_get_proto_info,
-		0,
+		METH_VARARGS,
 		_mysql_ConnectionObject_get_proto_info__doc__
 	},
 	{
 		"get_server_info",
 		(PyCFunction)_mysql_ConnectionObject_get_server_info,
-		0,
+		METH_VARARGS,
 		_mysql_ConnectionObject_get_server_info__doc__
 	},
 	{
 		"info",
 		(PyCFunction)_mysql_ConnectionObject_info,
-		0,
+		METH_VARARGS,
 		_mysql_ConnectionObject_info__doc__
 	},
 	{
 		"insert_id",
 		(PyCFunction)_mysql_ConnectionObject_insert_id,
-		0,
+		METH_VARARGS,
 		_mysql_ConnectionObject_insert_id__doc__
 	},
 	{
@@ -1676,7 +1948,7 @@ static PyMethodDef _mysql_ConnectionObject_methods[] = {
 	{
 		"ping",
 		(PyCFunction)_mysql_ConnectionObject_ping,
-		0,
+		METH_VARARGS,
 		_mysql_ConnectionObject_ping__doc__
 	},
 	{
@@ -1694,7 +1966,7 @@ static PyMethodDef _mysql_ConnectionObject_methods[] = {
 	{
 		"shutdown",
 		(PyCFunction)_mysql_ConnectionObject_shutdown,
-		0,
+		METH_VARARGS,
 		_mysql_ConnectionObject_shutdown__doc__
 	},
 	{
@@ -1706,13 +1978,13 @@ static PyMethodDef _mysql_ConnectionObject_methods[] = {
 	{
 		"store_result",
 		(PyCFunction)_mysql_ConnectionObject_store_result,
-		0,
+		METH_VARARGS,
 		_mysql_ConnectionObject_store_result__doc__
 	},
 	{
 		"string_literal",
 		(PyCFunction)_mysql_string_literal,
-		1,
+		METH_VARARGS,
 		_mysql_string_literal__doc__},
 	{
 		"thread_id",
@@ -1723,7 +1995,7 @@ static PyMethodDef _mysql_ConnectionObject_methods[] = {
 	{
 		"use_result",
 		(PyCFunction)_mysql_ConnectionObject_use_result,
-		0,
+		METH_VARARGS,
 		_mysql_ConnectionObject_use_result__doc__
 	},
 	{NULL,              NULL} /* sentinel */
@@ -1790,7 +2062,7 @@ static PyMethodDef _mysql_ResultObject_methods[] = {
 	{
 		"describe",
 		(PyCFunction)_mysql_ResultObject_describe,
-		0,
+		METH_VARARGS,
 		_mysql_ResultObject_describe__doc__
 	},
 	{
@@ -1830,7 +2102,7 @@ static MyMemberlist(_mysql_ResultObject_memberlist)[] = {
 		),
 	{NULL} /* Sentinel */
 };
-
+                                                                        
 static PyObject *
 _mysql_ConnectionObject_getattr(
 	_mysql_ConnectionObject *self,
@@ -1844,13 +2116,18 @@ _mysql_ConnectionObject_getattr(
 	PyErr_Clear();
 	if (strcmp(name, "closed") == 0)
 		return PyInt_FromLong((long)!(self->open));
-/*         if (strcmp(name, "__members__") == 0) */
-/*                 return listmembers(_mysql_ConnectionObject_memberlist); */
 #if PY_VERSION_HEX < 0x02020000
 	return PyMember_Get((char *)self, _mysql_ConnectionObject_memberlist, name);
 #else
-        PyErr_SetString(PyExc_AttributeError, name);
-        return NULL;	
+	{
+		MyMemberlist(*l);
+		for (l = _mysql_ConnectionObject_memberlist; l->name != NULL; l++) {
+			if (strcmp(l->name, name) == 0)
+				return PyMember_GetOne((char *)self, l);
+		}
+		PyErr_SetString(PyExc_AttributeError, name);
+		return NULL;
+	}
 #endif
 }
 
@@ -1866,12 +2143,17 @@ _mysql_ResultObject_getattr(
 		return res;
 	PyErr_Clear();
 #if PY_VERSION_HEX < 0x02020000
-/*         if (strcmp(name, "__members__") == 0) */
-/*                 return listmembers(_mysql_ResultObject_memberlist); */
 	return PyMember_Get((char *)self, _mysql_ResultObject_memberlist, name);
 #else
-        PyErr_SetString(PyExc_AttributeError, name);
-        return NULL;
+	{
+		MyMemberlist(*l);
+		for (l = _mysql_ResultObject_memberlist; l->name != NULL; l++) {
+			if (strcmp(l->name, name) == 0)
+				return PyMember_GetOne((char *)self, l);
+		}
+		PyErr_SetString(PyExc_AttributeError, name);
+		return NULL;
+	}
 #endif
 }
 
@@ -1926,11 +2208,7 @@ _mysql_ResultObject_setattr(
 }
 
 PyTypeObject _mysql_ConnectionObject_Type = {
-#ifndef MS_WIN32
-	PyObject_HEAD_INIT(&PyType_Type)
-#else
 	PyObject_HEAD_INIT(NULL)
-#endif
 	0,
 	"_mysql.connection", /* (char *)tp_name For printing */
 	sizeof(_mysql_ConnectionObject),
@@ -2002,9 +2280,9 @@ PyTypeObject _mysql_ConnectionObject_Type = {
 	0, /* (descrsetfunc) tp_descr_set */
 	0, /* (long) tp_dictoffset */
 	(initproc)_mysql_ConnectionObject_Initialize, /* tp_init */
-	(allocfunc)PyType_GenericAlloc, /* tp_alloc */
-	(newfunc)PyType_GenericNew, /* tp_new */
-	(destructor)_PyObject_GC_Del, /* tp_free Low-level free-memory routine */
+	NULL, /* tp_alloc */
+	NULL, /* tp_new */
+	NULL, /* tp_free Low-level free-memory routine */ 
 	0, /* (PyObject *) tp_bases */
 	0, /* (PyObject *) tp_mro method resolution order */
 	0, /* (PyObject *) tp_defined */
@@ -2014,11 +2292,7 @@ PyTypeObject _mysql_ConnectionObject_Type = {
 } ;
 
 PyTypeObject _mysql_ResultObject_Type = {
-#ifndef MS_WIN32
-	PyObject_HEAD_INIT(&PyType_Type)
-#else
 	PyObject_HEAD_INIT(NULL)
-#endif
 	0,
 	"_mysql.result",
 	sizeof(_mysql_ResultObject),
@@ -2091,9 +2365,9 @@ PyTypeObject _mysql_ResultObject_Type = {
 	0, /* (descrsetfunc) tp_descr_set */
 	0, /* (long) tp_dictoffset */
 	(initproc)_mysql_ResultObject_Initialize, /* tp_init */
-	(allocfunc)PyType_GenericAlloc, /* tp_alloc */
-	(newfunc)PyType_GenericNew, /* tp_new */
-	(destructor)_PyObject_GC_Del, /* tp_free Low-level free-memory routine */
+	NULL, /* tp_alloc */
+	NULL, /* tp_new */
+	NULL, /* tp_free Low-level free-memory routine */
 	0, /* (PyObject *) tp_bases */
 	0, /* (PyObject *) tp_mro method resolution order */
 	0, /* (PyObject *) tp_defined */
@@ -2149,18 +2423,32 @@ _mysql_methods[] = {
 	{
 		"get_client_info",
 		(PyCFunction)_mysql_get_client_info,
-		0,
+		METH_VARARGS,
 		_mysql_get_client_info__doc__
 	},
 #if MYSQL_VERSION_ID >= 32314
 	{
 		"thread_safe",
 		(PyCFunction)_mysql_thread_safe,
-		0,
+		METH_VARARGS,
 		_mysql_thread_safe__doc__
 	},
-	{NULL, NULL} /* sentinel */
 #endif
+#if MYSQL_VERSION_ID >= 40000
+	{
+		"server_init",
+		(PyCFunction)_mysql_server_init,
+		METH_VARARGS | METH_KEYWORDS,
+		_mysql_server_init__doc__
+	},
+	{
+		"server_end",
+		(PyCFunction)_mysql_server_end,
+		METH_VARARGS,
+		_mysql_server_end__doc__
+	},
+#endif
+	{NULL, NULL} /* sentinel */
 };
 
 static PyObject *
@@ -2197,10 +2485,18 @@ init_mysql(void)
 	PyObject *dict, *module, *emod, *edict;
 	module = Py_InitModule4("_mysql", _mysql_methods, _mysql___doc__,
 				(PyObject *)NULL, PYTHON_API_VERSION);
-#ifdef MS_WIN32
+	if (!module) return; /* this really should never happen */
 	_mysql_ConnectionObject_Type.ob_type = &PyType_Type;
 	_mysql_ResultObject_Type.ob_type = &PyType_Type;
+#if PY_VERSION_HEX >= 0x02020000
+	_mysql_ConnectionObject_Type.tp_alloc = PyType_GenericAlloc;
+	_mysql_ConnectionObject_Type.tp_new = PyType_GenericNew;
+	_mysql_ConnectionObject_Type.tp_free = _PyObject_GC_Del; 
+	_mysql_ResultObject_Type.tp_alloc = PyType_GenericAlloc;
+	_mysql_ResultObject_Type.tp_new = PyType_GenericNew;
+	_mysql_ResultObject_Type.tp_free = _PyObject_GC_Del;
 #endif
+
 	if (!(dict = PyModule_GetDict(module))) goto error;
 	if (PyDict_SetItemString(dict, "version_info",
 			       PyRun_String(version_info, Py_eval_input,
